// ModelSetupDialog.cpp : implementation file
//

#include "stdafx.h"
#include "sutra.h"
#include "ModelSetupDialog.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include "SUTRA_pie.h"



/////////////////////////////////////////////////////////////////////////////
// CModelSetupDialog dialog

#define SIZEOF_ARRAY(a) (sizeof(a) / sizeof(a[0]))
static UINT aUserFlags[] = { IDC_AREALMODEL, IDC_CROSSSECTIONAL, IDC_CYLINDRICAL, IDC_USERSPEC, IDC_ENERGY, IDC_FISHNET, IDC_IRREGULAR, IDC_SATUNSAT, IDC_SATUTATED, IDC_SOLUTECONST, IDC_SOLUTEVAR};
static UINT crossectionalFlags[] = { IDC_CYLINDRICAL, /*IDC_USERSPEC,*/ IDC_SATUNSAT, /*IDC_SATUTATED,*/ IDC_SOLUTEVAR, IDC_ENERGY};



CModelSetupDialog::CModelSetupDialog(CWnd* pParent /*=NULL*/)
	: CBaseDialog(CModelSetupDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(CModelSetupDialog)
	m_Orientation = 1;
	m_MeshDesign = 1;
	m_FlowConditions = 0;
	m_TransportConditions = 0;
	m_ModelThickness = 0;
	m_SpecifyModel = 0;
	//}}AFX_DATA_INIT

	m_SpecifyModel = 1;
	m_Orientation = 0;
	m_TransportConditions = 1;
}

CModelSetupDialog::~CModelSetupDialog()
{
}


void CModelSetupDialog::DoDataExchange(CDataExchange* pDX)
{
	CBaseDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CModelSetupDialog)
	DDX_Radio(pDX, IDC_AREALMODEL, m_Orientation);
	DDX_Radio(pDX, IDC_FISHNET, m_MeshDesign);
	DDX_Radio(pDX, IDC_SATUTATED, m_FlowConditions);
	DDX_Radio(pDX, IDC_SOLUTEVAR, m_TransportConditions);
	DDX_Radio(pDX, IDC_USERSPEC, m_ModelThickness);
	DDX_Radio(pDX, IDC_SPECIFYMODEL, m_SpecifyModel);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CModelSetupDialog, CBaseDialog)
	//{{AFX_MSG_MAP(CModelSetupDialog)
	ON_BN_CLICKED(IDC_SPECIFYMODEL, OnSpecifymodel)
	ON_BN_CLICKED(IDC_AREALMODEL, OnOrientation)
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDC_PROBLEM_SPECIFIC, OnSpecifymodel)
	ON_BN_CLICKED(IDC_CROSSSECTIONAL, OnOrientation)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CModelSetupDialog message handlers

void CModelSetupDialog::FillInfo(SUTRAProjectInfo* info)
{
	info->defaultSetting = !m_SpecifyModel;
	info->orientation = m_Orientation == 0? SUTRAProjectInfo::kAREAL : SUTRAProjectInfo::kCROSS_SECTIONAL;
	info->flowConditions = m_FlowConditions == 0 ? SUTRAProjectInfo::kSATURATED : SUTRAProjectInfo::kSAT_UNSAT;
	info->modelThickness = m_ModelThickness == 0? SUTRAProjectInfo::kUSER_SPEC : SUTRAProjectInfo::kCYLINDRICAL;
	info->meshDesign = m_MeshDesign == 0? SUTRAProjectInfo::kFISHNET_MESH : SUTRAProjectInfo::kIRREGULAR_MESH;

	switch (m_TransportConditions)
	{
		case 0:
				info->transportConditions = SUTRAProjectInfo::kSOLUTE_VAR;
			break;
		case 1:
				info->transportConditions = SUTRAProjectInfo::kSOLUTE_CONST;
			break;
		case 2:
				info->transportConditions = SUTRAProjectInfo::kENERGY;
	}
}

// this function taken from "Visual C++/MFC Frequently Asked Question"
void CModelSetupDialog::EnableControls(UINT* pControls, UINT cControls, BOOL fVisible)
{
    for (UINT uIndex = 0; uIndex < cControls; uIndex++)
    {
        CWnd* pwnd = GetDlgItem(pControls[uIndex]);
        if (pwnd)
        {
            pwnd->EnableWindow(fVisible);
        }
    }
}

void CModelSetupDialog::OnSpecifymodel() 
{
	// TODO: Add your control notification handler code here
	UpdateData(TRUE);

	if (m_SpecifyModel)
	{
	  EnableControls(aUserFlags, SIZEOF_ARRAY(aUserFlags), TRUE);

		if (0 == m_Orientation)
		  EnableControls(crossectionalFlags, SIZEOF_ARRAY(crossectionalFlags), FALSE);
		
	}
	else
	{
	  EnableControls(aUserFlags, SIZEOF_ARRAY(aUserFlags), FALSE);
	}
}

BOOL CModelSetupDialog::OnInitDialog() 
{
	CBaseDialog::OnInitDialog();
	
	EnableControls(aUserFlags, SIZEOF_ARRAY(aUserFlags), m_SpecifyModel);
	EnableControls(crossectionalFlags, SIZEOF_ARRAY(crossectionalFlags), FALSE);
		m_FlowConditions = 0;
		m_TransportConditions = 1;
		m_ModelThickness = 0;

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CModelSetupDialog::OnOrientation() 
{
	UpdateData(TRUE);

	if (m_Orientation == 0)
	{
	  EnableControls(crossectionalFlags, SIZEOF_ARRAY(crossectionalFlags), FALSE);
		m_FlowConditions = 0;
		m_TransportConditions = 1;
		m_ModelThickness = 0;
	}
	else
	{
	  EnableControls(crossectionalFlags, SIZEOF_ARRAY(crossectionalFlags), TRUE);
	}

	UpdateData(FALSE);
}
