#include "ANEPIE.h"
#include	"ImportPIE.h"
#include "ANECB.h"

class PeriodsInfo
{
	public:
		PeriodsInfo();
		~PeriodsInfo();

		void SetNumPeriods(int numPeriods);
		int GetNumPeriods() const;

		void SetNumVariables(int numvariables);
		int GetNumVariables() const;

		void SetVariableName(int variable, const char* name);
		CString GetVariableName(int variable) const;

		void SetAvailableVariable(int period, int variable, BOOL isAvailable);
		BOOL GetAvailableVariable(int period, int variable) const;

		void SetVariableTimeStep(int period, int variable);
		int GetVariableTimeStep(int period) const;

		void SetRequestedVariable(int period, int variable, BOOL isRequested);
		BOOL GetRequestedVariable(int period, int variable) const;

		BOOL	GetStatus() const {return m_status;}
	protected:
		void SetStatus(BOOL stat) {m_status = stat;}
	private:
		int	m_numPeriods;
		int m_numVariables;

		int* m_availabilityList;

		int* m_requestList;
		CString* m_namesList;

		BOOL	m_status;

		int m_timeStep;
		int* m_timeStepList;

};

class ifstream;
class CFortranFileParser
{
	public:
		CFortranFileParser();

	protected:
		char	line[255];
		long	lineIndex;

		void GetLine();
		long ReadI(long width = 0);
		double ReadG(long width = 0);
		void insert_E(char* num);
		void Skip(long width);

		void Open(char* fileName);

		BOOL GetStatus();
		void SetStatus(BOOL stat);

	private:
		ifstream* fortranFile;
		BOOL	status;		
};

class CParseSutraD5File : public CFortranFileParser
{
	public:
		CParseSutraD5File();
		~CParseSutraD5File();

		BOOL Parse(char* fileName);

		long GetNN();
		long GetNE();

		double GetX(long II);
		double GetY(long II);

		long GetIIN1(long LL);
		long GetIIN2(long LL);
		long GetIIN3(long LL);
		long GetIIN4(long LL);

		BOOL IsEnergyTransportSimulation();
		BOOL IsArealSimulation();

	private:

		// dataset 1
		char	SIMULA[12];

		// dataset 2
		char TITLE1[80];
		char TITLE2[80];

		// dataset 3
		long NN;
		long NE;
		long NBI;
		long NPINCH;
		long NPBC;
		long NUBC;
		long NSOP;
		long NSOU;
		long NOBS;
		long NTOBS;

		// dataset 4
		long IUNSAT;
		long ISSFLO;
		long ISSTRA;
		long IREAD;
		long ISTORE;
		
		// dataset 5
		double	UP;
		double GNU;

		// dataset 6
		long ITMAX;
		double DELT;
		double TMAX;
		long ITCYC;
		double DTMULT;
		double DTMAX;
		long NPCYC;
		long NUCYC;

		// dataset 7
		long NPRINT;
		long KNODAL;
		long KELMNT;
		long KINCID;
		long KPLOTP;
		long KPLOTU;
		long KVEL;
		long KBUDG;

		// dataset 8
		long ITRMAX;
		double RPMAX;
		double RUMAX;

		// dataset 9
		double COMPLF;
		double CW;
		double SIGMAW;
		double RHOW0;
		double URHOW0;
		double DRWDU;
		double VISC0;

		// dataset 10
		double COMPMA;
		double CS;
		double SIGMAS;
		double RHOS;

		// dataset 11
		char ADSMOD[10];
		double CHI1;
		double CHI2;

		// dataset 12
		double PRODF0;
		double PRODS0;
		double PRODF1;
		double PRODS1;

		//dataset 13
		double GRAVX;
		double GRAVY;

		// dataset 14A
		double SCALX;
		double SCALY;
		double SCALTH;
		double PORFAC;

		// dataset 14

		double* X;
		double* Y;
		double* THICK;
		double* POR;

		// dataset 15A
		double PMAXFA;
		double PMINDA;
		double ANGFAC;
		double ALMAXF;
		double ALMINF;
		double ATAVGF;

		// dataset 15B
		double* PMAX;
		double* PMIN;
		double* ANGLEX;
		double* ALMAX;
		double* ALMIN;
		double* ATAVG;

		// dataset 16
		long IDIREC;
		long NLINPI;
		long LCHAPI;
		long NCHAPL;

		double PBASE;

		double UBASE;

		// dataset 17
		long* IQCP;
		double* QINC;
		double* UINC;

		// dataset 18
		long* IQCU;
		double* QUINC;

		// dataset 19
		long* IPBC;
		double* PBC;
		double* UBC;

		// dataset 20
		long IUBC;
		double* CUBC;

		// dataset 21
		long NOBCYC;
		
		long* INOB;

		// dataset 22
		long* IIN1;
		long* IIN2;
		long* IIN3;
		long* IIN4;

		long* IEDGE1;
		long* IEDGE2;
		long* IEDGE3;
		long* IEDGE4;
};


class CParseSutraOutput : public CFortranFileParser
{
	public:
		CParseSutraOutput(BOOL D3File, BOOL temperatureFile, BOOL headFile);
		~CParseSutraOutput();

		BOOL ParseHeader(char*);

		long GetNumPeriods() const;
		long GetNumVariables() const;
		CString GetVariableName(long variable) const;
		BOOL IsVariableAvailable(long period, long variable) const;

		BOOL ParseAvailableVariables();
		double* GetVariableData(long period, long variable);
		long GetNumPoints() const;
		double GetPointX(long pointIndex) const;
		double GetPointY(long pointIndex) const;

		int* variableTimeStep;

	protected:
		BOOL SkipToHeader();

		BOOL SkipToNextPeriod();

	private:
		long numPeriods;
		long numVariables;
		CString*	variableNames;
		long*	variableAvailability;
		
		long numPoints;
		double* posX;
		double* posY;
		double** variableData;

		BOOL isD3File;
		BOOL isTemperatureFile;
		BOOL isHeadFile;
};

void
GSUTRAPostProc(ANE_PTR aneHandle, const ANE_STR fileName, ANE_PTR layerHandle);
