/******************************************************************************\
* 
* this module was developed by Argus Development Group
*
*  MODULE:      SUTRAPOSTPROC_PIE.C
*
*  PURPOSE:     To provide a Sutra Post Proc PIE template. 
*
*  First Version: December 1996 by Yaacov Ben-Israel.
*
*  COMMENTS:   
\******************************************************************************/
#include "stdafx.h"
#include <windows.h>
#include "sutrapostproc_pie.h"
#include <math.h>
#include <winuser.h>
#include <winbase.h>
#include <string.h>
#include <stdio.h>

#include <iostream.h>
#include <fstream.h>

#include "sutra.h"

#include "ANECB.h"

#include "SelectPostProcDialog.h"


////////////////////////////////////////////////////////////////////////
// PeriodsInfo class
////////////////////////////////////////////////////////////////////////
PeriodsInfo::PeriodsInfo()
{
	m_numPeriods = 0;
	m_numVariables = 0;

	m_availabilityList = 0;
	m_requestList = 0;
	m_namesList = 0;
	m_status = 1;
	m_timeStep = 0;
	m_timeStepList = 0;
}

PeriodsInfo::~PeriodsInfo()
{
	if (m_availabilityList)
		delete[] m_availabilityList;

	if (m_requestList)
		delete[] m_requestList;

	if (m_namesList)
		delete[] m_namesList;

	if (m_timeStepList)
		delete[] m_timeStepList;
}

void 
PeriodsInfo::SetNumPeriods(int numPeriods)
{
	m_numPeriods = numPeriods;

	if (m_availabilityList)
		delete[] m_availabilityList;

	if (m_requestList)
		delete[] m_requestList;

	m_availabilityList = new int[numPeriods];
	if (!m_availabilityList)
	{
		this->SetStatus(FALSE);
		return;
	}

	m_requestList = new int[numPeriods];
	if (!m_requestList)
	{
		this->SetStatus(FALSE);
		return;
	}

	m_timeStepList = new int[numPeriods];
	if (!m_timeStepList)
	{
		this->SetStatus(FALSE);
		return;
	}

	for (int i = 0; i < numPeriods; i++)
	{
		m_availabilityList[i] = 0;
		m_requestList[i] = 0;
		m_timeStepList[i] = 0;
	}
}

int 
PeriodsInfo::GetNumPeriods() const
{
	return m_numPeriods;
}

void 
PeriodsInfo::SetNumVariables(int numVariables)
{
	if (m_numVariables)
		delete[] m_namesList;

	m_numVariables = numVariables;
	m_namesList = new CString[numVariables];
	if (!m_namesList)
	{
		this->SetStatus(FALSE);
		return;
	}
}

int 
PeriodsInfo::GetNumVariables() const
{
	return m_numVariables;
}

void 
PeriodsInfo::SetVariableName(int variable, const char* name)
{
	ASSERT(variable >= 0 && variable < m_numVariables);

	if (variable >= 0 && variable < m_numVariables)
		m_namesList[variable]= name;
}

CString 
PeriodsInfo::GetVariableName(int variable) const
{
	ASSERT(variable >= 0 && variable < m_numVariables);

	if (variable >= 0 && variable < m_numVariables)
		return m_namesList[variable];
	else
		return "";
}

void 
PeriodsInfo::SetAvailableVariable(int period, int variable, BOOL isAvailabe)
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
	{
		if (isAvailabe)
			m_availabilityList[period] |= 1 << variable;
		else
			m_availabilityList[period] &= ~(1 << variable); 
	}
}

BOOL 
PeriodsInfo::GetAvailableVariable(int period, int variable) const
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
		return m_availabilityList[period] & (1 << variable) ? TRUE : FALSE;
	else
		return FALSE;
}

void 
PeriodsInfo::SetRequestedVariable(int period, int variable, BOOL isRequested)
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
	{
		if (isRequested)
			m_requestList[period] |= 1 << variable;
		else
			m_requestList[period] &= ~(1 << variable); 
	}
}

int 
PeriodsInfo::GetVariableTimeStep(int period) const
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
		return m_timeStepList[period];
	else
		return 0;
}

void 
PeriodsInfo::SetVariableTimeStep(int period, int variable)
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
	{
		m_timeStepList[period] = variable;
	}
}

BOOL 
PeriodsInfo::GetRequestedVariable(int period, int variable)const
{
	ASSERT(period >= 0);
	ASSERT(period < m_numPeriods);

	if (period >= 0 && period < m_numPeriods)
		return m_requestList[period] & (1 << variable) ? TRUE : FALSE;
	else
		return FALSE;
}


/////////////////////////////////////////////////////////////////////////////////


enum EDiagramType {kContoursDiagram, k3DChart, kColorMap, /*kPathLines,*/ kArrowsMap, kAlIOfThem};

void
GSUTRAPostProc(ANE_PTR aneHandle, const ANE_STR fileName, ANE_PTR layerHandle)
{

	ASSERT(0);
	BOOL cancel = FALSE;

//1) Locate result files (both directory and names).
//	Some models output fixed files, other can be customized by the input files, and yet others can be customized during run-time, usually by a dialog-box.

/*
	char VariablesFileName[255];	
	ANE_BOOL foundVariables =  ANE_FileGet(aneHandle, "Please select SUTRA Variables file:", 0, "*.D5", 0 , VariablesFileName, 255);
	if (!foundVariables)
	{	
		// the user choose 'Cancel'
		return;
	}	
*/
	char D5FileName[255];	
//gzh	ANE_BOOL foundD5 =  ANE_FileGet(aneHandle, "Please select D5 file:", 0, "*.D5", 0 , D5FileName, 255);
//gzh  The default name doesn't "work" yet, according to Shahar.  It will in future
//gzh  versions, in which case it should be fileName.D5?
	char defaultName[255];
	strcpy(defaultName,fileName);
	ANE_BOOL foundD5 =  ANE_FileGetEx(aneHandle, "Please select D5 file:", /*path*/0, defaultName, /*forexport*/0, D5FileName, 255, "*.D5", "SUTRA D5 file (*.D5)", 'TEXT');
	if (!foundD5)
	{	
		// the user choose 'Cancel'
		cancel = TRUE;
	}

	char D3FileName[255];
	char D4FileName[255];

	if (!cancel)
	{
		strcpy(D3FileName, D5FileName);
		D3FileName[strlen(D3FileName) - 1] = '3';

		strcpy(D4FileName, D5FileName);
		D4FileName[strlen(D4FileName) - 1] = '4';
	}

//2) For each result file: take only the relevant information from the file, and convert it to a format suitable for the Data Layer. In some cases, it may be necessary to combine several file, before the information can be extracted.

	CParseSutraD5File D5Parser;

	if (!cancel)
	{
		if (!D5Parser.Parse(D5FileName))
		{
			ANE_UserMessageOk(aneHandle, "Couldn't parse D5 file. Import aborted.");
			cancel = TRUE;
		}
	}

	BOOL isTemperatureFile;
	BOOL isHeadFile;
	
	if (!cancel)
	{
		isTemperatureFile = D5Parser.IsEnergyTransportSimulation();
		isHeadFile = D5Parser.IsArealSimulation();
	};

	//ASSERT(0);
	CParseSutraOutput D3Parser(TRUE, isTemperatureFile, isHeadFile); // D3 file

	if (!cancel)
	{
		if (!D3Parser.ParseHeader(D3FileName))
		{
			ANE_UserMessageOk(aneHandle, "Couldn't parse header of D3 file.");
			cancel = TRUE;
		}
	}

	CParseSutraOutput D4Parser(FALSE, isTemperatureFile, isHeadFile); // D4 file
	
	if (!cancel)
	{
		if (!D4Parser.ParseHeader(D4FileName))
		{
			ANE_UserMessageOk(aneHandle, "Couldn't parse header of D4 file.");
			cancel = TRUE;
		}
	}

	if (!cancel)
	{
		if (D3Parser.GetNumPeriods() != D4Parser.GetNumPeriods())
		{
			ANE_UserMessageOk(aneHandle, "Number of periods in D3 file differs from the number of periods in D4 file. Import aborted.");
			cancel = TRUE;
		}
	}

	PeriodsInfo periodsInfo;

	long numPeriods;
	
	if (!cancel)
	{
		numPeriods = D4Parser.GetNumPeriods();

		periodsInfo.SetNumPeriods(numPeriods);
		periodsInfo.SetNumVariables(4);

		if (isHeadFile)
			periodsInfo.SetVariableName(0, "Head");
		else
			periodsInfo.SetVariableName(0, "Pressure");
		if (isTemperatureFile)
			periodsInfo.SetVariableName(1, "Temperature");
		else
			periodsInfo.SetVariableName(1, "Concentration");
		periodsInfo.SetVariableName(2, "Saturation");
		periodsInfo.SetVariableName(3, "Velocity");

		for (long periodIndex = 0; periodIndex < numPeriods; periodIndex++)
		{
			periodsInfo.SetAvailableVariable(periodIndex, 0, D3Parser.IsVariableAvailable(periodIndex, 0));
			periodsInfo.SetAvailableVariable(periodIndex, 1, D3Parser.IsVariableAvailable(periodIndex, 1));
			periodsInfo.SetAvailableVariable(periodIndex, 2, TRUE);
			periodsInfo.SetAvailableVariable(periodIndex, 3, TRUE);
			periodsInfo.SetVariableTimeStep(periodIndex, D3Parser.variableTimeStep[periodIndex]);
		}
	}

	if (!cancel)
	{
		cancel = !periodsInfo.GetStatus();
	}

// 2.1) let the user select the type of Post-Processing objects
	
	BOOL overlayCharts;

	if (!cancel)
	{
		CSelectPostProcDialog	selectPostProcDialog(&periodsInfo);

		if (selectPostProcDialog.DoModal() != IDOK)
		{
			cancel = TRUE;
		}
		
		if (!cancel)
			overlayCharts = selectPostProcDialog.m_overlayCharts;
	}
	//ASSERT(0);

	if (!cancel)
	{
		if (!D3Parser.ParseAvailableVariables())
			{
			ANE_UserMessageOk(aneHandle, "Couldn't parse required variables from D3 file.");
			cancel = TRUE;
		}
	}

    
	if (!cancel)
	{
		if (D5Parser.GetNN() != D3Parser.GetNumPoints())
		{
			ANE_UserMessageOk(aneHandle, "Number of nodes in D5 files differs from the number of nodes in D3 file.");
			cancel = TRUE;
		}
	}


	if (!cancel)
	{
		if (!D4Parser.ParseAvailableVariables())
		{
			ANE_UserMessageOk(aneHandle, "Couldn't parse requested variables from D4 file.");
			cancel = TRUE;
		}
	}

	if (!cancel)
	{
		if (D5Parser.GetNE() != D4Parser.GetNumPoints())
		{
			ANE_UserMessageOk(aneHandle, "Number of elements in D5 files differs from the number of elements in D4 file.");
			cancel = TRUE;
		}
	}

//3) Create some new Data Layers, if not already exist, or clear existing layers.
	
	char* D3DataLayerName = "D3 Data";
	ANE_PTR D3DataLayerHandle = 0;

	if (!cancel)
	{
		D3DataLayerHandle = ANE_LayerGetHandleByName(aneHandle,D3DataLayerName);
		if (!D3DataLayerHandle)
		{
			char* D3DataLayerTemplate = 
		"Layer:\n\
		{\n\
		Name: \"D3 Data\"\n\
		Units: \"\"\n\
		Type: \"Data\"\n\
		Visible: No\n\
		Interpretation Method: Nearest\n\
	}\n";
		
			D3DataLayerHandle =  ANE_LayerAddByTemplate(aneHandle,  D3DataLayerTemplate, 0);
			if (!D3DataLayerHandle)
			{
				ANE_UserMessageOk(aneHandle, "problem creating new data layer for D3 info");
				cancel = TRUE;
			}
		}
}

	char* D4DataLayerName = "D4 Data";

	ANE_PTR D4DataLayerHandle = 0;

	if (!cancel)
	{
		D4DataLayerHandle = ANE_LayerGetHandleByName(aneHandle,D4DataLayerName);
		if (!D4DataLayerHandle)
		{
			char* D4DataLayerTemplate = 
		"Layer:\n\
		{\n\
		Name: \"D4 Data\"\n\
		Units: \"\"\n\
		Type: \"Data\"\n\
		Visible: No\n\
		Interpretation Method: Nearest\n\
	}\n";
		
			D4DataLayerHandle =  ANE_LayerAddByTemplate(aneHandle,  D4DataLayerTemplate, 0);
			if (!D4DataLayerHandle)
			{
				ANE_UserMessageOk(aneHandle, "problem creating new data layer for D4 info");
				cancel = TRUE;
			}
		}
	}
//4) Each converted data file, should be read into a Data Layer
	double minx, miny, maxx, maxy;

	ANE_INT32 D3numPoints = 0;
	ANE_DOUBLE* D3posX = 0;

	if (!cancel)
	{
		D3numPoints = D5Parser.GetNN();
		D3posX = new ANE_DOUBLE[D3numPoints]; 
		if (!D3posX)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d point for D3. Import aborted", D3numPoints);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	ANE_DOUBLE* D3posY = 0;
	
	if (!cancel)
	{
		D3posY = new ANE_DOUBLE[D3numPoints]; 
		if (!D3posY)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d point for D3. Import aborted", D3numPoints);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	if (!cancel)
	{
		for (int i = 0; i < D3numPoints; i++)
		{
			D3posX[i] = D5Parser.GetX(i+1);
			D3posY[i] = D5Parser.GetY(i+1);

			if (i == 0)
			{
				minx = maxx = D3posX[i];
				miny = maxy = D3posY[i];
			}
			else
			{
				minx = min(minx, D3posX[i]);
				miny = min(miny, D3posY[i]);
				maxx = max(maxx, D3posX[i]);
				maxy = max(maxy, D3posY[i]);
			}
		}
	}


	ANE_INT32 numD3DataParameters=0;
	ANE_DOUBLE** D3DataParameters = 0;
	
	if (!cancel)
	{
		D3DataParameters = new ANE_DOUBLE*[numPeriods*3];

		if (!D3DataParameters)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d time periods for D3. Import aborted", numPeriods);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	ANE_STR* D3ParamNames = 0;
	
	if (!cancel)
	{
		D3ParamNames = new ANE_STR[numPeriods*3];
		if (!D3ParamNames)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d time periods for D3. Import aborted", numPeriods);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
		else
		{
			for (long tempIndex = 0; tempIndex < numPeriods*3; tempIndex++)
			{
				D3ParamNames[tempIndex] = 0;
			}
		}
	}

	//ASSERT(0);
	if (!cancel)
	{
		bool datafound = FALSE;
		for (long D3PeriodIndex = 0; D3PeriodIndex < numPeriods; D3PeriodIndex++)
		{
			for (long variableIndex = 0; variableIndex < 4; variableIndex++)
				if (periodsInfo.GetRequestedVariable(D3PeriodIndex, variableIndex))
				{
					datafound = TRUE;

					if (variableIndex < 3) 
					{

						D3DataParameters[numD3DataParameters] = D3Parser.GetVariableData(D3PeriodIndex, variableIndex);
			
						char buf[200];
						sprintf(buf, "Period %d %s", periodsInfo.GetVariableTimeStep(D3PeriodIndex), (const char*)D3Parser.GetVariableName(variableIndex));

						char* varName = new char[strlen(buf)+1];
						if (!varName)
						{
							char buffer[500];
							sprintf(buffer, "Out of memory while allocating names for D3. Import aborted");
							ANE_UserMessageOk(aneHandle, buffer);
							cancel = TRUE;
						}

						if (!cancel)
						{
							strcpy(varName, buf);
							D3ParamNames[numD3DataParameters] = varName;
							numD3DataParameters++;
						}

						if (!cancel)
						{
							//ASSERT(0);
							ANE_DOUBLE mindata;
							ANE_DOUBLE maxdata;
							ANE_DOUBLE* temp = *D3DataParameters;
							mindata = maxdata = *temp;
							for (long dataValueIndex = 0; dataValueIndex < D3numPoints; dataValueIndex++)
							{
								mindata = min(mindata, *temp);
								maxdata = max(maxdata, *temp);
								temp++;
							}
							if (mindata == maxdata)
							{
								char buffer[500];
								sprintf(buffer, "D3 Data is uniform. Import aborted");
								ANE_UserMessageOk(aneHandle, buffer);
								cancel = TRUE;
							}
						}
					}
				}
		}
		cancel = !datafound || cancel;
	}
	//ASSERT(0);


	ANE_INT32 numTriangles = 0;
	ANE_INT32* node0 = 0;
	ANE_INT32* node1 = 0;
	ANE_INT32* node2 = 0;

	if (!cancel)
	{
		numTriangles = 2*D5Parser.GetNE();
		node0 = new ANE_INT32[numTriangles];
		node1 = new ANE_INT32[numTriangles];
		node2 = new ANE_INT32[numTriangles];
		if (!node0 || !node1 || !node2)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating mesh geometry for %d elements. Import aborted", numTriangles);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	if (!cancel)
	{
		for (int ll = 0; ll < numTriangles/2; ll++)
		{
			node0[ll*2] = D5Parser.GetIIN1(ll+1)-1;
			node1[ll*2] = D5Parser.GetIIN2(ll+1)-1;
			node2[ll*2] = D5Parser.GetIIN3(ll+1)-1;

			node0[ll*2+1] = D5Parser.GetIIN3(ll+1)-1;
			node1[ll*2+1] = D5Parser.GetIIN4(ll+1)-1;
			node2[ll*2+1] = D5Parser.GetIIN1(ll+1)-1;
		}
 
		 ANE_DataLayerSetTriangulatedData(aneHandle, 
							D3DataLayerHandle, 
							D3numPoints,
							D3posX,
							D3posY,
							numTriangles,
							node0,
							node1,
							node2,
							numD3DataParameters,
							D3DataParameters,
							D3ParamNames);
	}

	if (D3posX)
	{
		delete D3posX;
		D3posX = 0;
	}

	if (D3posY)
	{
		delete D3posY;
		D3posY = 0;
	}

	if (D3DataParameters)
	{
		delete[] D3DataParameters;
		D3DataParameters = 0;
	}

	for (long i = 0; i < numD3DataParameters; i++)
	{
	 // DONT call this:	delete[] D3DataParameters[i]; it is deleted in the destructor of the parser
		if (D3ParamNames[i])
		{
			delete[] D3ParamNames[i];
			D3ParamNames[i] = 0;
		}
	}

	delete[] D3DataParameters;
	delete[] D3ParamNames;

	if (node0)
	{
		delete node0;
		node0 = 0;
	}

	if (node1)
	{
		delete node1;
		node1 = 0;
	}

	if (node2)
	{
		delete node2;
		node2 = 0;
	}


	/////////////////////////////
	//  repeat this step for the elements (velocity) parameters
	/////////////////////////////

	long D4numPoints = 0;
	double* D4posx = 0;
	double* D4posy = 0;

	if (!cancel)
	{
		D4numPoints = D4Parser.GetNumPoints();
		D4posx = new double[D4numPoints];
		D4posy = new double[D4numPoints];

		if (!D4posx || !D4posy)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d points for D4. Import aborted", D4numPoints);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	if (!cancel)
	{
		for (long D4PointIndex = 0; D4PointIndex < D4numPoints; D4PointIndex++)
		{
			D4posx[D4PointIndex] = D4Parser.GetPointX(D4PointIndex);
			D4posy[D4PointIndex] = D4Parser.GetPointY(D4PointIndex);
		}
	}

	ANE_INT32 numD4DataParameters=0;
	ANE_DOUBLE** D4DataParameters = 0;
	ANE_STR* D4ParamNames = 0;

	if (!cancel)
	{
		D4DataParameters = new ANE_DOUBLE*[numPeriods*2];
		D4ParamNames = new ANE_STR[numPeriods*2];

		if (!D4DataParameters || !D4ParamNames)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating %d time periods for D4. Import aborted", numPeriods);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}		

	if (!cancel)
	{
		for (long D4PeriodIndex = 0; D4PeriodIndex < numPeriods; D4PeriodIndex++)
			if (periodsInfo.GetRequestedVariable(D4PeriodIndex, 3))
			{
				for (long D4VarIndex = 0; D4VarIndex < 2; D4VarIndex++)
				{
					D4DataParameters[numD4DataParameters] = D4Parser.GetVariableData(D4PeriodIndex, D4VarIndex);

					char buf[200];
					sprintf(buf, "Period %d %s", periodsInfo.GetVariableTimeStep(D4PeriodIndex), (const char*)D4Parser.GetVariableName(D4VarIndex));

					char* varName = new char[strlen(buf)+1];
					if (!varName)
					{
						char buffer[500];
						sprintf(buffer, "Out of memory while allocating names for D4. Import aborted");
						ANE_UserMessageOk(aneHandle, buffer);
						return;
					}
					
					strcpy(varName, buf);
					D4ParamNames[numD4DataParameters] = varName;
					numD4DataParameters++;
				}
			}
	}

	if (!cancel)
	{
	 ANE_DataLayerSetData(aneHandle, 
						D4DataLayerHandle, 
						D4numPoints,
						D4posx,
						D4posy,
						numD4DataParameters,
						D4DataParameters,
						D4ParamNames);
	}

	if (D4posx)
	{
		delete[] D4posx;
		D4posx = 0;
	}

	if (D4posy)
	{
		delete[] D4posy;
		D4posy = 0;
	}

	for (i = 0; i < numD4DataParameters; i++)
	{
	 // DONT call this:	delete[] D4DataParameters[i]; it is delted in the destructor of the parser
		delete[] D4ParamNames[i];
	}

	if (D4DataParameters)
	{
		delete[] D4DataParameters;
		D4DataParameters = 0;
	}

	if (D4ParamNames)
	{
		delete[] D4ParamNames;
		D4ParamNames = 0;
	}

//5) Create some new Maps Layer, or clear existing ones






	char* SUTRAMapsLayerName = "SUTRA Post Processing Charts";
	
	ANE_PTR SUTRAMapsLayerHandle = 0;

	if (!cancel)
	{
		SUTRAMapsLayerHandle = ANE_LayerGetHandleByName(aneHandle,SUTRAMapsLayerName);
		if (!SUTRAMapsLayerHandle)
		{
	char* SUTRAMapsLayerTemplate =
	"Layer:\n"
	"{\n"
	"	Name: \"SUTRA Post Processing Charts\"\n"
	"	Units: \"\"\n"
	"	Type: \"Maps\"\n"
	"	Visible: Yes\n"
	"}\n";

			SUTRAMapsLayerHandle =  ANE_LayerAddByTemplate(aneHandle,  SUTRAMapsLayerTemplate, 0);
			if (!SUTRAMapsLayerName)
			{
				ANE_UserMessageOk(aneHandle, "problem creating new maps layer");
				cancel = 0;
			}
		}
		else
		{
			// clear previous charts
			if (ANE_UserMessageOkCancel(aneHandle, "The Maps Layer \"SUTRA Post Processing Charts\" already exist. Do you want to clear its content? Click Ok to clear, Cancel to add new Charts to layer."))
				ANE_LayerClear(aneHandle,  SUTRAMapsLayerHandle, 0 /* false */);
		}
	}
//6) Add some Post-Processing object in each Maps Layer.
	 

	static char* contourCharString = 
	"\n\
Contours Chart:{\n\
	SOURCE LAYER: \"D3 Data\"\n\
	BOUND RECT:{\n\
		Origin:\n\
		{\n\
		x: %.15g\n\
		y: %.15g\n\
		}\n\
		Extent:\n\
		{\n\
			x: %.15g\n\
			y: %.15g\n\
		}\n\
	}\n\
	VALUE: \"%s\"\n\
	Contouring Algorithm: Linear\n\
	Overlay Source Data: %d\n\
	Show Head Title: %d\n\
	Head Title: \"%s\"\n\
	Show Legend: %d\n\
	Draw Axes: %d\n\
}\n\
";

	static char* arrowsCharString = 
	"\n\
Arrows Chart:{\n\
	SOURCE LAYER: \"D4 Data\"\n\
	BOUND RECT:{\n\
		Origin:\n\
		{\n\
		x: %.15g\n\
		y: %.15g\n\
		}\n\
		Extent:\n\
		{\n\
			x: %.15g\n\
			y: %.15g\n\
		}\n\
	}\n\
	X VALUE: \"%s\"\n\
	Y VALUE: \"%s\"\n\
	Overlay Source Data: %d\n\
	Show Head Title: %d\n\
	Head Title: \"%s\"\n\
	Show Legend: %d\n\
	Draw Axes: %d\n\
	Color By Length: 0\n\
	Thickness: 1\n\
}\n\
";

	long numDataParameters = 0;
	long chartIndex = 0;
		int placeIndex = 0;

	char* PDLString = 0;

	if (!cancel)
	{
		numDataParameters = numD3DataParameters + numD4DataParameters/2;
		
		long bufferLengthPerChart = max(strlen(contourCharString), strlen(arrowsCharString)) + 800;

		PDLString = new char[bufferLengthPerChart*numDataParameters];
		if (!PDLString)
		{
			char buffer[500];
			sprintf(buffer, "Out of memory while allocating preparing post processing objects. Import aborted", numPeriods);
			ANE_UserMessageOk(aneHandle, buffer);
			cancel = TRUE;
		}
	}

	long numChartsInLine = 0;
	long numChartsInCol = 0;
	double extentx = 0.0;
	double extenty = 0.0;
	double orix = 0.0;
	double oriy = 0.0;

	if (!cancel)
	{
		PDLString[0] = '*';
		PDLString[1] = '\0';
		
		numChartsInLine = (long) ceil(sqrt(numDataParameters));
		numChartsInCol = numChartsInLine;
		//extent = 0.9 * min(maxx-minx, maxy-miny)/numChartsInLine;
		extentx = 0.9 * (maxx-minx)/numChartsInLine;
		extenty = 0.9 * (maxy-miny)/numChartsInLine;
		orix = 0.0;
		oriy = 0.0;
 
		int showTitle = 1;

		for (long periodIndex = 0; periodIndex < numPeriods; periodIndex++)
		{
			for (long variableIndex = 0; variableIndex < 3; variableIndex++)
			{
				if (periodsInfo.GetRequestedVariable(periodIndex, variableIndex))
				{
					char buf[200];
					sprintf(buf, "Period %d %s", periodsInfo.GetVariableTimeStep(periodIndex), (const char*)D3Parser.GetVariableName(variableIndex));
					
					orix = minx + chartIndex%numChartsInLine * extentx/0.9;
					oriy = miny + chartIndex/numChartsInLine * extenty/0.9;

					sprintf(PDLString+strlen(PDLString),  contourCharString, orix, oriy, extentx, extenty, buf, overlayCharts, showTitle, buf, showTitle, showTitle);
					placeIndex = strlen(PDLString);

					chartIndex++;
					if (overlayCharts)
						showTitle=0;
				}
			}

			if (periodsInfo.GetRequestedVariable(periodIndex, 3))
			{
				char bufx[200];
				sprintf(bufx, "Period %d %s", periodsInfo.GetVariableTimeStep(periodIndex), (const char*)D4Parser.GetVariableName(0));
				char bufy[200];
				sprintf(bufy, "Period %d %s", periodsInfo.GetVariableTimeStep(periodIndex), (const char*)D4Parser.GetVariableName(1));
				char buftitle[200];
				//sprintf(buftitle, "Period %d Velocity", periodIndex+1);
				sprintf(buftitle, "Period %d Velocity", periodsInfo.GetVariableTimeStep(periodIndex));
				
				orix = minx + chartIndex%numChartsInLine * extentx/0.9;
				oriy = miny + chartIndex/numChartsInLine * extenty/0.9;

				sprintf(PDLString+strlen(PDLString),  arrowsCharString, orix, oriy, extentx, extenty, bufx, bufy, overlayCharts, showTitle, buftitle, showTitle, showTitle);
				placeIndex = strlen(PDLString);

				chartIndex++;
				if (overlayCharts)
					showTitle=0;
			}
		}
	}
	
	if (!cancel)
	{
		ANE_ImportTextToLayerByHandle(aneHandle, SUTRAMapsLayerHandle, PDLString);
	}

	if (PDLString)
	{
		delete[] PDLString;
		PDLString = 0;
	} 
}




//=================================================================
//	PIE handling functions and declerations
//=================================================================
/*
struct ANEPIEDesc* gFunDesc[20];								// list of PIE descriptors for all parts

//
// "SUTRA" project part

struct ANEPIEDesc 					gSUTRAPPImportPIEDesc;			// PIE defcriptor
struct ImportPIEDesc				gSUTRAPPPIPIEDesc;		// ImportPIE descriptor

void
GetANEFunctions(ANE_INT32& numNames, struct ANEPIEDesc**& descriptors)
{ 
	numNames = 0;

	//
	//
	// prepare PROJECT descriptor for SUTRA project
	
	gSUTRAPPPIPIEDesc.version = IMPORT_PIE_VERSION;
	gSUTRAPPPIPIEDesc.name = "Sutra Post Proc...";						// name of project
	gSUTRAPPPIPIEDesc.importFlags = (EImportPIEFlags)kImportAllwaysVisible;
	gSUTRAPPPIPIEDesc.toLayerTypes = kPIEAnyLayer; // was kPIETriMeshLayer
	// gSUTRAPPPIPIEDesc.fromLayerTypes = ;	//not used
	gSUTRAPPPIPIEDesc.doImportProc = GSUTRAPostProc;// address of Post Processing Function function
	// gSUTRAPPPIPIEDesc.neededProject = ;			// not used

	//
	// prepare PIE descriptor for "SUTRA" project

	gSUTRAPPImportPIEDesc.name = "Sutra Post Proc...";			// PIE name
	gSUTRAPPImportPIEDesc.type = kImportPIE;					// PIE type: project PIE
	gSUTRAPPImportPIEDesc.descriptor = &gSUTRAPPPIPIEDesc;	// pointer to descriptor
	
	gFunDesc[(numNames)++] = &gSUTRAPPImportPIEDesc;			// add descriptor to list



	descriptors = gFunDesc;
}
*/
///////////////////////////////////////////////////////////////////
/// Fortran File Parser
///////////////////////////////////////////////////////////////////

CFortranFileParser::CFortranFileParser()
{
	fortranFile = 0;

}

//CFortranFileParser::~CFortranFileParser()
//{
//	if (fortranFile)
//		delete(fortranFile);
//}

 
void 	CFortranFileParser::GetLine()
{
	ASSERT(fortranFile);

	if (this->GetStatus())
	{
		fortranFile->getline(line, 255);
		lineIndex = 0;
	}
}

long 
CFortranFileParser::ReadI(long width /* = 0*/)
{
	long result;
	long numRead;
	if (width > 0)
	{
		char tmp[80];
		strncpy(tmp, line+lineIndex, width);
		tmp[width] = '\0';
		numRead = sscanf(tmp, "%ld", &result);
		if (numRead == 1)
			lineIndex+=width;
	}
	else
	{
		long pos;
		numRead = sscanf(line+lineIndex, "%ld%n", &result, &pos);
		
		if (numRead == 1)
			lineIndex += pos;
	}
	
	if (numRead != 1)
		this->SetStatus(FALSE);
	return result;
}

double 
CFortranFileParser::ReadG(long width /* = 0 */)
{	
	double result;
	long numRead;
    char tmp[80];
	char* dLocation;

	if (width > 0)
	{
		strncpy(tmp, line+lineIndex, width);

		dLocation = strchr(tmp, 'D');
		if (dLocation)
			*dLocation = 'E';

		tmp[width] = '\0';
		insert_E(tmp);
		numRead = sscanf(tmp, "%lf", &result);
		lineIndex+=width;
	}
	else
	{
		char *cp;
		char *cpns;
		long pos;
	    char c;

		pos = 0;
		cpns = line+lineIndex;
		while ( *cpns == ' ' ) 
		{
		  cpns++;
		  pos++;
		}
		cp = strchr(cpns, ' ');
        if ( cp != NULL )
		{
			c = *cp;
			*cp = '\0';
		}
	    strcpy(tmp, cpns);
        if ( (dLocation = strchr(tmp, 'D')) != NULL )
			*dLocation = 'E';
		pos += strlen(tmp);
		insert_E(tmp);
		if ( cp != NULL )
		   *cp = c;
			
		numRead = sscanf(tmp, "%lf", &result);
		
		if (numRead == 1)
			  lineIndex += pos;
		
    }
	if (numRead != 1)
		this->SetStatus(FALSE);
	return result;
}

void
CFortranFileParser::insert_E(char* num)
{
	char *cp;
    int i, end, len;

	cp = num + strlen(num) - 1;
	while ( cp != num && *cp != '+' && *cp != '-' )
		cp--;
	if ( cp != num && isdigit(*(cp-1)) )
	{
       end = strlen(num) + 1;
	   len = strlen(cp) + 1;
	   for (i=0; i< len; i++ )
		   *(num+end-i) = *(num+end-i-1);
	   *cp = 'E';
	}
}

void 
CFortranFileParser::Skip(long width)
{
	lineIndex+=width;
}

void
CFortranFileParser::Open(char* fileName)
{
	fortranFile = new ifstream(fileName);
	status = fortranFile && fortranFile->good();
}

BOOL 
CFortranFileParser::GetStatus()
{
	if (fortranFile && fortranFile->good() || fortranFile->eof())
		return status;
	else
		return FALSE;
}

void 
CFortranFileParser::SetStatus(BOOL stat)
{
	status = stat;
}



CParseSutraD5File::CParseSutraD5File()
{

		// dataset 14

	X = 0;
	Y = 0;
	THICK = 0;
	POR = 0;

		// dataset 15B
	PMAX = 0;
	PMIN = 0;
	ANGLEX = 0;
	ALMAX = 0;
	ALMIN = 0;
	ATAVG = 0;

		// dataset 17
	IQCP = 0;
	QINC = 0;
	UINC = 0;

		// dataset 18
	IQCU = 0;
	QUINC = 0;

		// dataset 19
	IPBC = 0;
	PBC = 0;
	UBC = 0;
	// dataset 20
	CUBC = 0;
		// dataset 21
	INOB = 0;

		// dataset 22
	IIN1 = 0;
	IIN2 = 0;
	IIN3 = 0;
	IIN4 = 0;

	IEDGE1 = 0;
	IEDGE2 = 0;
	IEDGE3 = 0;
	IEDGE4 = 0;
}


CParseSutraD5File::~CParseSutraD5File()
{
		// dataset 14

		if (X)
			delete X;

		if (Y)
			delete Y;

		if(THICK)
			delete THICK;
		if(POR)
			delete POR;

		// dataset 15B
		if(PMAX)
			delete PMAX;
		if(PMIN)
			delete PMIN;
		if(ANGLEX)
			delete ANGLEX;
		if(ALMAX)
			delete ALMAX;
		if(ALMIN)
			delete ALMIN;
		if(ATAVG)
			delete ATAVG;

		// dataset 17
		if (IQCP)
			delete IQCP;
		if (QINC)
			delete QINC;
		if (UINC)
			delete UINC;

		// dataset 18
		if (IQCU)
			delete IQCU;
		if (QUINC)
			delete QUINC;

		// dataset 19
		if (IPBC)
			delete IPBC;
		if (PBC)
			delete PBC;
		if (UBC)
			delete UBC;


		// dataset 21
		if (INOB)
			delete INOB;

		// dataset 22
		if (IIN1)
			delete IIN1;
		if (IIN2)
			delete IIN2;
		if (IIN3)
			delete IIN3;
		if (IIN4)
			delete IIN4;

		if (IEDGE1)
			delete IEDGE1;
		if (IEDGE2)
			delete IEDGE2;
		if (IEDGE3)
			delete IEDGE3;
		if (IEDGE4)
			delete IEDGE4;
}


BOOL	CParseSutraD5File::Parse(char* fileName)
{

	this->Open(fileName);

	// dataset1
	this->GetLine();
	strncpy(SIMULA, line, 12);


	// dataset 2
	this->GetLine();
	strncpy(TITLE1, line, 80);

	this->GetLine();
	strncpy(TITLE2, line, 80);

	// dataset 3
	this->GetLine();

	NN = ReadI(5);
	NE = ReadI(5);
	NBI = ReadI(5);
	NPINCH = ReadI(5);
	NPBC = ReadI(5);
	NUBC = ReadI(5);
	NSOP = ReadI(5);
	NSOU = ReadI(5);
	NOBS = ReadI(5);
	NTOBS = ReadI(5);

	// dataset 4
	this->GetLine();
	IUNSAT = ReadI(5);
	ISSFLO = ReadI(5);
	ISSTRA = ReadI(5);
	IREAD = ReadI(5);
	ISTORE = ReadI(5);
		
	// dataset 5
	this->GetLine();
	UP = ReadG(10);
	GNU  = ReadG(15);

	// dataset 6
	this->GetLine();
	ITMAX = ReadI(5);
	DELT = ReadG(15);
	TMAX = ReadG(15);
	ITCYC = ReadI(10);
	DTMULT = ReadG(10);
	DTMAX = ReadG(15);
	NPCYC = ReadI(5);
	NUCYC = ReadI(5);

	// dataset 7
	this->GetLine();
	NPRINT = ReadI(5);
	KNODAL = ReadI(5);
	KELMNT = ReadI(5);
	KINCID = ReadI(5);
	KPLOTP = ReadI(5);
	KPLOTU = ReadI(5);
	KVEL = ReadI(5);
	KBUDG = ReadI(5);

	// dataset 8
	this->GetLine();
	ITRMAX = ReadI(10);
	RPMAX = ReadG(10);
	RUMAX = ReadG(10);

	// dataset 9
	this->GetLine();
	COMPLF = ReadG(10);
	CW = ReadG(10);
	SIGMAW = ReadG(10);
	RHOW0 = ReadG(10);
	URHOW0 = ReadG(10);
	DRWDU = ReadG(10);
	VISC0 = ReadG(10);

		// dataset 10
	this->GetLine();
	COMPMA = ReadG(10);
	CS = ReadG(10);
	SIGMAS = ReadG(10);
	RHOS = ReadG(10);

		// dataset 11

	this->GetLine();
	strncpy(ADSMOD, line, 10);
	lineIndex+=10;
	CHI1 = ReadG(10);
	CHI2 = ReadG(10);

		// dataset 12
	this->GetLine();
	PRODF0 = ReadG(10);
	PRODS0 = ReadG(10);
	PRODF1 = ReadG(10);
	PRODS1 = ReadG(10);

		//dataset 13
	this->GetLine();
	GRAVX = ReadG(10);
	GRAVY = ReadG(10);

		// dataset 14A

	this->GetLine();
	this->Skip(5);
	this->Skip(5);
	SCALX = ReadG(10);
	SCALY = ReadG(10);
	SCALTH = ReadG(10);
	PORFAC = ReadG(10);

		// dataset 14

	X = new double[NN];
	Y = new double[NN];

	if (!X || !Y)
	{
		this->SetStatus(FALSE);
		return FALSE;
	}
	// THICK = new double[NN];
	// POR = new double[NN];
	long nodeIndex;
	for (nodeIndex = 0; this->GetStatus() && nodeIndex < NN; nodeIndex++)
	{
		this->GetLine();

		long II = ReadI(5);
		ASSERT(II==nodeIndex+1);

		if (II != nodeIndex+1)
			this->SetStatus(FALSE);

//		long IUNSATREG = ReadI(5);
		this->Skip(5);

		X[nodeIndex] = ReadG(10);
		Y[nodeIndex] = ReadG(10);
		// THICK[nodeIndex] = ReadG(10);
		// POR[nodeIndex] = ReadG(10);
		
	}
	// dataset 15A
	this->GetLine();
	this->Skip(10);

	PMAXFA = ReadG(10);
	PMINDA = ReadG(10);
	ANGFAC = ReadG(10);
	ALMAXF = ReadG(10);
	ALMINF = ReadG(10);
	ATAVGF = ReadG(10);

		// dataset 15B

	//PMAX= new double[NE];
	//PMIN= new double[NE];
	//ANGLEX= new double[NE];
	//ALMAX= new double[NE];
	//ALMIN= new double[NE];
	//ATAVG= new double[NE];
	long elemIndex;
	for (elemIndex = 0; this->GetStatus() && elemIndex < NE; elemIndex++)
	{
		this->GetLine();

		long L = ReadI(10);
		ASSERT(L==elemIndex+1);
	
		if (L != elemIndex + 1)
			this->SetStatus(FALSE);

		this->Skip(5); //?

		//PMAX[elemIndex] = ReadG(10);
		//PMIN[elemIndex] = ReadG(10);
		//ANGLEX[elemIndex] = ReadG(10);
		//ALMAX[elemIndex] = ReadG(10);
		//ALMIN[elemIndex] = ReadG(10);
		//ATAVG[elemIndex] = ReadG(10);
		
	}
	
	// dataset 16

	if (KPLOTP || KPLOTU)
	{
		this->GetLine();
		IDIREC = ReadI(5);
		NLINPI = ReadI(5);
		LCHAPI = ReadI(5);
		NCHAPL = ReadI(5);

		if (KPLOTP)
			PBASE = ReadG(13);
		
		if (KPLOTU)
			UBASE = ReadG(13);
	}
		// dataset 17

	if (NSOP > 0)
	{
		for (long i = 0; this->GetStatus() && i < NSOP; i++)
		{
			this->GetLine();
			//	long* IQCP; (I10)
			//  double* QINC; (G15)
			//  double* UINC; (G15)
		}
		// read the extra line
		this->GetLine();
	}
	// dataset 18

	if (NSOU)
	{
		for (long i = 0;this->GetStatus() && i < NSOU; i++)
		{
			this->GetLine();
		//	long* IQCU; (I10)
		//	double* QUINC; (G15)
		}
		this->GetLine();
	}

	// dataset 19
	if (NPBC)
	{
		for (long i = 0; this->GetStatus() && i < NPBC; i++)
		{
			this->GetLine();
	//	long* IPBC; (I5)
	//	double* PBC; (G20)
	//	double* UBC; (G20)
		}
		this->GetLine();
	}
	// dataset 20
	if (NUBC)
	{
		for (long i = 0; this->GetStatus() && i < NUBC; i++)
		{
			this->GetLine();
		//	long IUBC; (I5)
		//	double* CUBC; (G20)
		}
		this->GetLine();
	}

		// dataset 21
	if (NOBS > 0)
	{
		this->GetLine();
		NOBCYC = ReadI(10);
		for (long i = 0; this->GetStatus() && i < (NOBS+16)/16; i++)
		{
			this->GetLine();
		 // long* INOB; (16I5)
		} 
	}
	// dataset 22
	IIN1 = new long[NE];
	IIN2 = new long[NE];
	IIN3 = new long[NE];
	IIN4 = new long[NE];
	if (!IIN1 || !IIN2 || !IIN3 || !IIN4)
	{
		this->SetStatus(FALSE);
		return FALSE;
	}

		//IEDGE1 = new long[NE];
		//IEDGE2 = new long[NE];
		//IEDGE3 = new long[NE];
		//IEDGE4 = new long[NE];
	for (elemIndex = 0; this->GetStatus() && elemIndex < NE; elemIndex++)
	{
		this->GetLine();
		long LL = ReadI(6);
		ASSERT(LL == elemIndex+1 || -LL == elemIndex+1);

		if (LL != elemIndex+1 && -LL != elemIndex+1)
			this->SetStatus(FALSE);

		IIN1[elemIndex] = ReadI(6);
		ASSERT(IIN1[elemIndex] > 0 && IIN1[elemIndex] <= NN);
		if (!(IIN1[elemIndex] > 0 && IIN1[elemIndex] <= NN))
			this->SetStatus(FALSE);

		IIN2[elemIndex] = ReadI(6);
		ASSERT(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN);
		if (!(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN))
			this->SetStatus(FALSE);

		IIN3[elemIndex] = ReadI(6);
		ASSERT(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN);
		if (!(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN))
			this->SetStatus(FALSE);

		IIN4[elemIndex] = ReadI(6);
		ASSERT(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN);
		if (!(IIN2[elemIndex] > 0 && IIN2[elemIndex] <= NN))
			this->SetStatus(FALSE);

		if (LL < 0)
		{
			this->GetLine();
		//	IEDGE1[elemIndex] = ReadI(6);
		//	IEDGE2[elemIndex] = ReadI(6);
		//	IEDGE3[elemIndex] = ReadI(6);
		//	IEDGE4[elemIndex] = ReadI(6);
		}
	}

	return this->GetStatus();
}


long CParseSutraD5File::GetNN()
{
	return NN;
}

long CParseSutraD5File::GetNE()
{
	return NE;
}

double CParseSutraD5File::GetX(long II)
{
	ASSERT(X);
	ASSERT(II>0 && II<=NN);

	if (X && II>0 && II<=NN)
		return X[II-1];
	else
	{
		this->SetStatus(FALSE);
		return 0.0;
	}
}

double  CParseSutraD5File::GetY(long II)
{
	ASSERT(Y);
	ASSERT(II>0 && II<=NN);

	if (Y && II>0 && II<=NN)
		return Y[II-1];
	else
	{
		this->SetStatus(FALSE);
		return 0.0;
	}
}

long CParseSutraD5File::GetIIN1(long LL)
{
	ASSERT(IIN1);
	ASSERT(LL>0 && LL<=NE);

	if (IIN1 && LL>0 && LL<=NE)
		return IIN1[LL-1];
	else
	{
		this->SetStatus(FALSE);
		return 0;
	}
}

long CParseSutraD5File::GetIIN2(long LL)
{
	ASSERT(IIN2);
	ASSERT(LL>0 && LL<=NE);

	if (IIN2 && LL>0 && LL<=NE)
		return IIN2[LL-1];
	else
	{
		this->SetStatus(FALSE);
		return 0;
	}
}

long CParseSutraD5File::GetIIN3(long LL)
{
	ASSERT(IIN3);
	ASSERT(LL>0 && LL<=NE);

	if (IIN3 && LL > 0 && LL <= NE)
		return IIN3[LL-1];
	else
	{
		this->SetStatus(FALSE);
		return 0;
	}
}

long CParseSutraD5File::GetIIN4(long LL){
	ASSERT(IIN4);
	ASSERT(LL>0 && LL<=NE);

	if (IIN4 && LL>0 && LL<=NE)
		return IIN4[LL-1];
	else
	{
		this->SetStatus(FALSE);
		return 0;
	}
}

BOOL CParseSutraD5File::IsEnergyTransportSimulation(){
	return SIMULA == "SUTRA ENERGY";
}

BOOL CParseSutraD5File::IsArealSimulation(){
	return (GRAVX == 0 && GRAVY == 0);
}

CParseSutraOutput::CParseSutraOutput(BOOL D3File, BOOL temperatureFile, BOOL headFile)
{
	numPeriods = 0;
	variableNames = 0;
	variableAvailability = 0;
	numPoints = 0;
	variableData = 0;
	posX = 0;
	posY = 0;
	variableTimeStep = 0;

	isD3File = D3File;
	isTemperatureFile = temperatureFile;
	isHeadFile = headFile;
}

CParseSutraOutput::~CParseSutraOutput()
{
	if (variableNames)
		delete[] variableNames;

	if (variableAvailability)
		delete[] variableAvailability;

	if (variableData)
	{
		for (int i = 0; i < numPeriods*numVariables; i++)
			if (variableData[i])
				delete[] variableData[i];
		delete[] variableData;
	}

	if (posX)
		delete[] posX;
	if (posY)
		delete[] posY;

	if (variableTimeStep)
		delete[] variableTimeStep;
}


BOOL 
CParseSutraOutput::SkipToHeader()
{
// the D3 header looks like this:

//## Argus Numerical Environments PostProcessing Sample                              
//## A test case with multiple time steps                                            
//## Nodewise Results
//##     936 nodes
//##    3 Time Steps printed
//##    Current Time   Calculated
//##      in Seconds   Press Temp  Time step
//##      1.0000D+01   1     1     1
//##      2.0000D+01   1     1     2
//##      3.0000D+01   1     1     3

// in this function we skip till the line: ##    3 Time Steps printed
	this->GetLine();
	this->GetLine();
	this->GetLine();
	this->GetLine();

	// this line is: ##     936 nodes
	ASSERT(!strncmp(line, "##", 2));	
	this->Skip(2);	
	numPoints = this->ReadI(); // read number of nodes

	this->GetLine();

	return this->GetStatus();
}

BOOL 
CParseSutraOutput::ParseHeader(char* fileName)
{
	this->Open(fileName);

	if (!this->SkipToHeader())
		return FALSE;

	// next line is: ##    1 Time Steps printed
	ASSERT(!strncmp(line, "##", 2));	

	if (strncmp(line, "##", 2))
	{
		this->SetStatus(FALSE);
		return FALSE;
	}

	this->Skip(2);	
	numPeriods = this->ReadI();

	if (isD3File)
	{
		// this is a D3 (node head/pressure, concentration/temperature and saturation) file

		this->GetLine();
		this->GetLine();

		numVariables = 3;
	
		variableNames = new CString[numVariables];
		if (!variableNames)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}

		if (isHeadFile)
			variableNames[0] = "Head";
		else
			variableNames[0] = "Pressure";
		if (isTemperatureFile)
			variableNames[1] = "Temperature";
		else
			variableNames[1] = "Concentration";
		variableNames[2] = "Saturation";

		variableAvailability = new long[numPeriods];
		if (!variableAvailability)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}

		variableTimeStep = new int[numPeriods];
		if (!variableAvailability)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}

		for (long periodIndex = 0; periodIndex < numPeriods; periodIndex++)
		{
			this->GetLine();
			ASSERT(!strncmp(line, "##", 2));	
			if (strncmp(line, "##", 2))
			{
				this->SetStatus(FALSE);
				return FALSE;
			}
		
			this->Skip(2);
			double timeStepSize = this->ReadG();

			variableAvailability[periodIndex] = 1 << 2; // initialize availability (saturation is always available)
			for (int i = 0; i < 2; i++)
			{
				// read availability of pressure and temperature
				BOOL available = this->ReadI();			
				variableAvailability[periodIndex] |= (available << i);
			}
			// set time step
			int timeStep = this->ReadI();
			variableTimeStep[periodIndex] = timeStep;
		}
	}
	else 
	{
		// this is a D4 (element velocity) file

		numVariables = 2;
	
		variableNames = new CString[numVariables];
		if (!variableNames)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}
		variableNames[0] = "XVelocity";
		variableNames[1] = "YVelocity";

		variableAvailability = new long[numPeriods];
		if (!variableAvailability)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}
		
		variableTimeStep = new int[numPeriods];
		if (!variableAvailability)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}

		for (long periodIndex = 0; periodIndex < numPeriods; periodIndex++)
		{
			variableAvailability[periodIndex] = 3; // velocity is always available
		}
	}
	return this->GetStatus();
}

long CParseSutraOutput::GetNumPeriods() const
{
	return numPeriods;
}

long CParseSutraOutput::GetNumVariables() const
{
	return numVariables;
}

CString CParseSutraOutput::GetVariableName(long variable) const
{
	ASSERT(variableNames);
	ASSERT(variable >= 0 && variable < numVariables);

	if (variableNames && variable >= 0 && variable < numVariables)
		return variableNames[variable];
	else
	{
		((CParseSutraOutput*)this)->SetStatus(FALSE);
		return "";
	}
}

BOOL CParseSutraOutput::IsVariableAvailable(long period, long variable) const
{
	ASSERT(variableAvailability);
	ASSERT(period >= 0 && period < numPeriods);

	if (variableAvailability && period >= 0 && period < numPeriods)
		return (variableAvailability[period] & (1 << variable)) != 0;
	else
	{
		((CParseSutraOutput*)this)->SetStatus(FALSE);
		return FALSE;
	}
}

BOOL 
CParseSutraOutput::SkipToNextPeriod()
{
	while(line[0] == '#' && this->GetStatus())
		this->GetLine();

	return GetStatus();
}


BOOL 
CParseSutraOutput::ParseAvailableVariables()
{
	variableData = new double*[numPeriods*numVariables];
	if (!variableData)
	{
		this->SetStatus(FALSE);
		return FALSE;
	}

	for (long i = 0; i < numPeriods*numVariables; i++)
		variableData[i] = 0;

	for (long periodIndex = 0; periodIndex < numPeriods; periodIndex++)
	{
		if (!this->SkipToNextPeriod())
			return FALSE;

		BOOL firstVariable = FALSE;

		///int readNumPoints = this->ReadI();
		///if (numPoints == 0)
		///{
		///	numPoints=readNumPoints;
		posX = new double[numPoints];
		posY = new double[numPoints];
		if (!posX || !posY)
		{
			this->SetStatus(FALSE);
			return FALSE;
		}
		firstVariable = TRUE;
		///}
		///
		///ASSERT(numPoints == readNumPoints);
		///if (numPoints != readNumPoints)
		///{
		///	this->SetStatus(FALSE);
		///	return FALSE;
		///}
		///int numParams = this->ReadI();

		int numParams = 2; // D4 file, don't count the x y coordinates
		if (isD3File)
			numParams = 3; // don't count the x y coordinates

		if (variableAvailability[periodIndex])
		{
			long variableIndex;
			for (variableIndex = 0; variableIndex< numVariables; variableIndex++)
			{
					if (this->IsVariableAvailable(periodIndex, variableIndex))
					{
						variableData[periodIndex*numVariables+variableIndex] = new double[numPoints];
						if (!variableData[periodIndex*numVariables+variableIndex])
						{
							this->SetStatus(FALSE);
							return FALSE;
						}
					} 
			}


			for (long pointIndex = 0; pointIndex < numPoints; pointIndex++)
			{
				if (pointIndex > 0)	// we are already on the first line of data
					this->GetLine();

				///if (isD3File)
				///{
				long Point = this->ReadI();

				ASSERT(Point == pointIndex+1);				
				if (Point != pointIndex+1)
				{
					this->SetStatus(FALSE);
					return FALSE;
				}
				///}

				double X = this->ReadG();
				double Y = this->ReadG();

				ASSERT(firstVariable || X == posX[pointIndex]); 
				ASSERT(firstVariable || Y == posY[pointIndex]); 


				if (firstVariable)
				{
					posX[pointIndex] = X;
					posY[pointIndex] = Y;
				}

				for (variableIndex = 0; variableIndex< numParams; variableIndex++)
				{
					double val = ReadG();
					if (this->IsVariableAvailable(periodIndex, variableIndex))
					{
						variableData[periodIndex*numVariables+variableIndex][pointIndex]  = val;
					} 
				}

				///if (!isD3File)
				///{
				///	long Point = this->ReadI();
				///	ASSERT(Point == pointIndex+1);					
				///	if (Point != pointIndex+1)
				///	{
				///		this->SetStatus(FALSE);
				///		return FALSE;
				///	}
				///}

			}
		}
		else
		{
			for (long i = 0; i < numPoints; i++)
				this->GetLine();
		}
		
		if (periodIndex != numPeriods-1)
		{
			this->GetLine();
			this->GetLine();
		}

	}
	
	return this->GetStatus();
}

long CParseSutraOutput::GetNumPoints() const
{
	ASSERT(numPoints > 0);
	
	if (numPoints > 0)
		return numPoints;
	else
	{
		((CParseSutraOutput*)this)->SetStatus(FALSE);
		return 0;
	}
}

double CParseSutraOutput::GetPointX(long pointIndex) const
{
	ASSERT(pointIndex >= 0 && pointIndex < numPoints);

	if (pointIndex >= 0 && pointIndex < numPoints)
		return posX[pointIndex];
	else
	{
		((CParseSutraOutput*)this)->SetStatus(FALSE);
		return 0.0;
	}
}

double CParseSutraOutput::GetPointY(long pointIndex) const
{
	ASSERT(pointIndex >= 0 && pointIndex < numPoints);

	if (pointIndex >= 0 && pointIndex < numPoints)
		return posY[pointIndex];
	else
	{
		((CParseSutraOutput*)this)->SetStatus(FALSE);
		return 0.0;
	}
}


double* CParseSutraOutput::GetVariableData(long period, long variable)
{
	ASSERT(variableData);
	ASSERT(period >= 0 && period < numPeriods);
	ASSERT(variable >= 0 && variable < numVariables);

	if (variableData && period >= 0 && period < numPeriods && variable >= 0 && variable < numVariables)
		return variableData[period*numVariables+variable];
	else
	{
		this->SetStatus(FALSE);
		return FALSE;
	}
}
